#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  [Item("ReplaceBranchMoveEvaluator", "")]
  [StorableClass]
  public class ReplaceBranchMoveEvaluator : SingleSuccessorOperator, ISymbolicRegressionMoveEvaluator {

    public override bool CanChangeName {
      get { return false; }
    }

    public ILookupParameter<ReplaceBranchMove> MoveParameter {
      get { return (ILookupParameter<ReplaceBranchMove>)Parameters["ReplaceBranchMove"]; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }

    public ILookupParameter<DoubleValue> MoveQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveQuality"]; }
    }

    public ILookupParameter<ISymbolicRegressionSingleObjectiveEvaluator> EvaluatorParameter {
      get { return (ILookupParameter<ISymbolicRegressionSingleObjectiveEvaluator>)Parameters["Evaluator"]; }
    }
    public ILookupParameter<IRegressionProblemData> ProblemDataParameter {
      get { return (ILookupParameter<IRegressionProblemData>)Parameters["ProblemData"]; }
    }
    public ILookupParameter<IntRange> FitnessCalculationPartitionParameter {
      get { return (ILookupParameter<IntRange>)Parameters["FitnessCalculationPartition"]; }
    }

    public ILookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (ILookupParameter<ISymbolicExpressionTree>)Parameters["SymbolicExpressionTree"]; }
    }
    [StorableConstructor]
    protected ReplaceBranchMoveEvaluator(bool deserializing) : base(deserializing) { }
    protected ReplaceBranchMoveEvaluator(ReplaceBranchMoveEvaluator original, Cloner cloner) : base(original, cloner) { }
    public ReplaceBranchMoveEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The quality of a symbolic regression solution."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveQuality", "The evaluated quality of a move on a symbolic regression solution."));
      Parameters.Add(new LookupParameter<ISymbolicRegressionSingleObjectiveEvaluator>("Evaluator", ""));
      Parameters.Add(new LookupParameter<ReplaceBranchMove>("ReplaceBranchMove", ""));
      Parameters.Add(new LookupParameter<IRegressionProblemData>("ProblemData", ""));
      Parameters.Add(new LookupParameter<IntRange>("FitnessCalculationPartition", ""));
      Parameters.Add(new LookupParameter<ISymbolicExpressionTree>("SymbolicExpressionTree", ""));
    }

    public override IOperation Apply() {
      var evaluator = EvaluatorParameter.ActualValue;
      var move = (ReplaceBranchMove)MoveParameter.ActualValue.Clone();

      move.Parent.RemoveSubtree(move.SubtreeIndex);
      move.Parent.InsertSubtree(move.SubtreeIndex, move.NewBranch);

      var problemData = ProblemDataParameter.ActualValue;
      IExecutionContext childContext = (IExecutionContext)ExecutionContext.CreateChildOperation(evaluator);

      var fitnessPartition = FitnessCalculationPartitionParameter.ActualValue;
      var rows = Enumerable.Range(fitnessPartition.Start, fitnessPartition.End - fitnessPartition.Start);
      var quality = evaluator.Evaluate(childContext, move.Tree, problemData, rows);

      MoveQualityParameter.ActualValue = new DoubleValue(quality);
      return base.Apply();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ReplaceBranchMoveEvaluator(this, cloner);
    }
  }
}
